/*-
 * ~~~~~~licensing~~~~~~
 * blueprint-generator
 * ==========
 * Copyright (C) 2020 - 2021 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.esb.system.management.blueprint.generator;

import freemarker.cache.URLTemplateLoader;
import freemarker.template.Configuration;
import freemarker.template.DefaultObjectWrapper;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.osgi.framework.BundleContext;
import ru.entaxy.esb.system.common.exception.TemplateNotFound;

import java.io.IOException;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Map;

import static freemarker.template.Configuration.SQUARE_BRACKET_INTERPOLATION_SYNTAX;

public class BlueprintGeneratorImpl implements BlueprintGenerator {

    private static final Log LOG = LogFactory.getLog(BlueprintGeneratorImpl.class.getName());
    private static final String SYSTEM_NAME = "systemName";
    private static final String FTL_EXTENSION = ".ftl";
    private static final String XML_EXTENSION = ".xml";

    private BundleContext bundleContext;

    @Override
    public Blueprint createBlueprint(URL templateUrl, String templateName, String systemName, Map<String, String> params)
            throws TemplateNotFound, TemplateException, IOException {
        return writeBlueprint(templateUrl, templateName, systemName, params);
    }

    @Override
    public Blueprint createBlueprint(String templateName, String systemName, Map<String, String> params) throws TemplateNotFound,
            TemplateException, IOException {
        return createBlueprint(bundleContext.getBundle().getEntry("/template/"), templateName, systemName, params);
    }

    private Blueprint writeBlueprint(URL templateUrl, String templateName, String systemName, Map<String, String> params)
            throws IOException, TemplateException, TemplateNotFound {
        Template temp = getTemplateByName(templateUrl, templateName);
        addSystemParam(params, systemName);
        try (StringWriter writer = new StringWriter()) {
            temp.process(params, writer);
            return new Blueprint(templateName + "-" + systemName + XML_EXTENSION, writer.toString().getBytes());
        }
    }

    private void addSystemParam(Map<String, String> params, String systemName) {
        params.put(SYSTEM_NAME, systemName);
    }

    private Template getTemplateByName(URL templateUrl, String templateName) throws IOException, TemplateNotFound {
        Configuration cfg = getConfiguration(templateUrl);
        return cfg.getTemplate(templateName + FTL_EXTENSION);
    }

    private Configuration getConfiguration(URL templateUrl) throws TemplateNotFound {
        Configuration freemarkerConfig = new Configuration();
        freemarkerConfig.setObjectWrapper(new DefaultObjectWrapper());
        freemarkerConfig.setLocalizedLookup(false);
        freemarkerConfig.setTemplateLoader(getURLTemplateLoader(templateUrl));
        freemarkerConfig.setInterpolationSyntax(SQUARE_BRACKET_INTERPOLATION_SYNTAX);
        return freemarkerConfig;
    }

    private URLTemplateLoader getURLTemplateLoader(URL templateUrl) {
        return new URLTemplateLoader() {
            @Override
            protected URL getURL(String templateName) {
                try {
                    return new URL(templateUrl + templateName);
                } catch (MalformedURLException e) {
                    LOG.error(e);
                }
                return null;
            }
        };
    }

    public void setBundleContext(BundleContext bundleContext) {
        this.bundleContext = bundleContext;
    }
}
